#include "DAC53401.h"
#include <stdio.h>
#include <string.h>


static IfxI2c_I2c g_i2cHandle;           /* I2C handle                                       */
IfxI2c_I2c_Device g_i2cDev1_DAC53401;    /* I2C Slave device handle to EEPROM of DAC53401 1     */
IfxI2c_I2c_Device g_i2cDev2_DAC53401;    /* I2C Slave device handle to EEPROM of DAC53401 2     */


/********************************************************************************************************************
 * Function Name : init_I2C_Module
 * Description   : I2c控制器初始化以及主从mcu的i2c设备初始化
 * Arguments     :
 * Return Value  :
 * Author        : yinjun
 * Data          : 2022.9.16
 ********************************************************************************************************************/
void init_I2C_Module(void)
{
    /* Initialize module */
    IfxI2c_I2c_Config i2cConfig;                     /* Create configuration structure               */
    IfxI2c_I2c_initConfig(&i2cConfig, &MODULE_I2C0); /* Fill structure with default values and Module
                                                        address                                      */
    /* I2c pin configuration */
    const IfxI2c_Pins DAC53401_PINS =
        {
            &DAC53401_SCL_PIN,           /* SCL port pin                                 */
            &DAC53401_SDA_PIN,           /* SDA port pin                                 */
            IfxPort_PadDriver_ttlSpeed1 /* Pad driver mode                              */
        };
    i2cConfig.pins = &DAC53401_PINS;    /* Configure port pins                          */
    i2cConfig.baudrate = I2C_BAUDRATE; /* Configure baud rate with 400kHz              */

    IfxI2c_I2c_initModule(&g_i2cHandle, &i2cConfig); /* Initialize module                            */

    /* Initialize device */
    IfxI2c_I2c_deviceConfig i2cDevice1Config;                     /* Create device configuration                  */
    IfxI2c_I2c_deviceConfig i2cDevice2Config;                     /* Create device configuration                  */
    IfxI2c_I2c_initDeviceConfig(&i2cDevice1Config, &g_i2cHandle); /* Fill structure with default values and I2C
                                                                    Handler                                      */
    IfxI2c_I2c_initDeviceConfig(&i2cDevice2Config, &g_i2cHandle); /* Fill structure with default values and I2C
                                                                        Handler                                      */
    /* Because it is 7 bit long and bit 0 is R/W bit, the device address has to be shifted by 1 */
    i2cDevice1Config.deviceAddress = DAC53401_ADDR0 << 1;
    IfxI2c_I2c_initDevice(&g_i2cDev1_DAC53401, &i2cDevice1Config); /* Initialize the I2C device handle             */

    i2cDevice2Config.deviceAddress = DAC53401_ADDR1 << 1;
    IfxI2c_I2c_initDevice(&g_i2cDev2_DAC53401, &i2cDevice2Config); /* Initialize the I2C device handle             */
}


/********************************************************************************************************************
 * Function Name : DAC53401_Reg_Init
 * Description   : 初始化DAC53401的寄存器
 * Arguments     : IfxI2c_I2c_Device i2c_dev：需要配置的i2c设备
 * Return Value  :
 * Author        : yinjun
 * Data          : 2022.9.16
 ********************************************************************************************************************/
void DAC53401_Reg_Init(IfxI2c_I2c_Device i2c_dev)
{
    uint8 readData[2];
#ifdef DEBUG
    uint8 databuf[2];
#endif
    uint8 addr[1];

    /* 软复位 */
    addr[0] = RESET_REG_ADDRESS;
    while(IfxI2c_I2c_write(&i2c_dev, addr, 1) == IfxI2c_I2c_Status_nak);
    while(IfxI2c_I2c_read(&i2c_dev, readData, 2) == IfxI2c_I2c_Status_nak);

#ifdef DEBUG
    Ifx_print("dataH = %x\r\n", readData[0]);
    Ifx_print("dataL = %x\r\n", readData[1]);
#endif

    /* 在原有配置的基础上，将软复位对应的位进行配置 */
    readData[1] = readData[1] | 0x0C;
    write_Reg(RESET_REG_ADDRESS, readData, i2c_dev);

#ifdef DEBUG
    while(IfxI2c_I2c_write(&g_i2cDevEeprom, addr, 1) == IfxI2c_I2c_Status_nak);
    while(IfxI2c_I2c_read(&g_i2cDevEeprom, databuf, 2) == IfxI2c_I2c_Status_nak);
    Ifx_print("dataH = %x\r\n", databuf[0]);
    Ifx_print("dataL = %x\r\n", databuf[1]);
#endif

    /* 设置基准电压为电源内部 */
    addr[0] = GENERAL_CONFIG_REG;

    while(IfxI2c_I2c_write(&i2c_dev, addr, 1) == IfxI2c_I2c_Status_nak);
    while(IfxI2c_I2c_read(&i2c_dev, readData, 2) == IfxI2c_I2c_Status_nak);

#ifdef DEBUG
    Ifx_print("dataH = %x\r\n", readData[0]);
    Ifx_print("dataL = %x\r\n", readData[1]);
#endif

    /* EN_PMBUS位置1、设置DAC_PDN位为00，也就是POWER UP模式、设置基准电压为内部参考电压1.21V，并设置放大倍数为4 */
    readData[0] = readData[0]  | (1 << 4);
    readData[1] = readData[1] | (1 << 2) | (3 << 0);
    readData[1] &= ~(1 << 4);
    write_Reg(GENERAL_CONFIG_REG, readData, i2c_dev);

#ifdef DEBUG
    while(IfxI2c_I2c_write(&g_i2cDevEeprom, addr, 1) == IfxI2c_I2c_Status_nak);
    while(IfxI2c_I2c_read(&g_i2cDevEeprom, databuf, 2) == IfxI2c_I2c_Status_nak);
    Ifx_print("dataH = %x\r\n", databuf[0]);
    Ifx_print("dataL = %x\r\n", databuf[1]);
#endif
}

/********************************************************************************************************************
 * Function Name : write_Reg
 * Description   : 向指定的寄存器当中写入数据
 * Arguments     : uint8 reg：寄存器地址
 *                 uint8 *writeData: 需要写入的数据
 *                 IfxI2c_I2c_Device i2c_dev: 需要写入数据的i2c设备
 * Return Value  :
 * Author        : yinjun
 * Data          : 2022.9.16
 ********************************************************************************************************************/
void write_Reg(uint8 reg, uint8 *writeData, IfxI2c_I2c_Device i2c_dev)
{
    uint8 databuf[3];
    memcpy(&databuf[0], &reg, 1);
    memcpy(&databuf[1], writeData, 2);

    while(IfxI2c_I2c_write(&i2c_dev, databuf, 3) == IfxI2c_I2c_Status_nak);
}


/********************************************************************************************************************
 * Function Name : read_Reg
 * Description   : 读取指定寄存器当中的数据
 * Arguments     : uint8 reg：寄存器地址
 *                 IfxI2c_I2c_Device i2c_dev: 需要读取写入数据的i2c设备
 * Return Value  : 读取到的数据
 * Author        : yinjun
 * Data          : 2022.9.16
 ********************************************************************************************************************/
uint16 read_Reg(uint8 reg, IfxI2c_I2c_Device i2c_dev)
{
    uint8 databuf[2];
    uint16 ret;

    while(IfxI2c_I2c_write(&i2c_dev, &reg, 1) == IfxI2c_I2c_Status_nak);    /* 向i2c设备写入需要读取的地址 */
    while(IfxI2c_I2c_read(&i2c_dev, databuf, 2) == IfxI2c_I2c_Status_nak);  /* 讲读取到的数据放到databuf中 */

    ret = databuf[0] << 8 | databuf[1];

    return ret;

#ifdef DEBUG
    Ifx_print("dataH = %x\r\n", databuf[0]);
    Ifx_print("dataL = %x\r\n", databuf[1]);
#endif
}

/********************************************************************************************************************
 * Function Name : tztek_DAC53401_Init
 * Description   : 初始化DAC53401，需要在主函数中调用
 * Arguments     :
 * Return Value  :
 * Author        : yinjun
 * Data          : 2022.9.16
 ********************************************************************************************************************/
void tztek_DAC53401_Init(void)
{
    init_I2C_Module();
    waitTime(IfxStm_getTicksFromMilliseconds(BSP_DEFAULT_TIMER, 30)); /* 等待DCA53401复位 */
    DAC53401_Reg_Init(g_i2cDev1_DAC53401);
    DAC53401_Reg_Init(g_i2cDev2_DAC53401);
}

/********************************************************************************************************************
 * Function Name : tztek_dac_Test
 * Description   : 向DAC53401数据寄存器中写入数据，输出一定的值
 * Arguments     :
 * Return Value  :
 * Author        : yinjun
 * Data          : 2022.9.16
 ********************************************************************************************************************/
void tztek_dac_Test(void)
{
    /*
    uint8 *data：写入数据寄存器DAC_DATA[9:0]当中的值，DAC_DATA为16位寄存器，实际值只取[11:2]，所以数据有效范围为4-4092
                Vout = DAC_DATA / 2^N * Vref * GAIN(输出计算公式，DAC_DATA为数据寄存器当中的值；N为10；Vref为1.21；GAIN为4)
                example: uint8 data[2]; data[0] = 0x0f; data[1] = 0xf0;  只取[11:2]位，所以实际取值为1111 1111 00，换算成十进
                                         制为1020，根据公式可以计算出预期输出为1020 / 1024 * 4.84 = 4.821V
    */
    uint8 V_out_1[2]= {0xff,0xff};  //4.821V
    uint8 V_out_2[2]= {0x0A,0xA0};  //3.214V
    static int cnt = 0;
    static int cnt2 = 0;
    
    if (cnt == 60000)
    {
        //输出4.821V 100次
        if (cnt2 < 100)
        {
            write_Reg(DATA_REGISTER_ADDR, V_out_1, g_i2cDev1_DAC53401);
            write_Reg(DATA_REGISTER_ADDR, V_out_1, g_i2cDev2_DAC53401);
        }
        //输出3.214V 100次
        else if (cnt2 == 100)
        {
//            Ifx_Console_print("MCU_A DAC1 OUTPUT: 3.21V\r\n");
//            Ifx_Console_print("MCU_A DAC2 OUTPUT: 3.21V\r\n");
        }
        else if (cnt2 < 200)
        {
            write_Reg(DATA_REGISTER_ADDR, V_out_2, g_i2cDev1_DAC53401);
            write_Reg(DATA_REGISTER_ADDR, V_out_2, g_i2cDev2_DAC53401);
            
        }
        else
        {
//            Ifx_Console_print("MCU_A DAC1 OUTPUT: 4.82V\r\n");
//            Ifx_Console_print("MCU_A DAC2 OUTPUT: 4.82V\r\n");
            cnt2 = 0;
        }


#ifdef DEBUG
        uint16 readData;
        uint8 databuf[2];
        readData = read_Reg(0x21, g_i2cDev1_DAC53401);
        databuf[0] = (uint8)(readData >> 8);
        databuf[1] = (uint8)readData;
        Ifx_print("dataH = %x\r\n", databuf[0]);
        Ifx_print("dataL = %x\r\n", databuf[1]);
#endif
        cnt2++;
        cnt = 0;
    }
    cnt++;
}
