/**********************************************************************************************************************
 * \file SPI_CPU.c
 * \copyright Copyright (C) Infineon Technologies AG 2019
 *
 * Use of this file is subject to the terms of use agreed between (i) you or the company in which ordinary course of
 * business you are acting and (ii) Infineon Technologies AG or its licensees. If and as long as no such terms of use
 * are agreed, use of this file is subject to following:
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization obtaining a copy of the software and
 * accompanying documentation covered by this license (the "Software") to use, reproduce, display, distribute, execute,
 * and transmit the Software, and to prepare derivative works of the Software, and to permit third-parties to whom the
 * Software is furnished to do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including the above license grant, this restriction
 * and the following disclaimer, must be included in all copies of the Software, in whole or in part, and all
 * derivative works of the Software, unless such copies or derivative works are solely in the form of
 * machine-executable object code generated by a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *********************************************************************************************************************/

/*********************************************************************************************************************/
/*-----------------------------------------------------Includes------------------------------------------------------*/
/*********************************************************************************************************************/
#include "tztek_tja1145.h"
#include "IfxPort.h"
#include "Ifx_Console.h"
#include "Bsp.h"
#include "IfxQspi_SpiMaster.h"


/*********************************************************************************************************************/
/*------------------------------------------------------Macros-------------------------------------------------------*/
/*********************************************************************************************************************/
/* QSPI modules */
#define QSPI3_MASTER                &MODULE_QSPI3   /* SPI Master module                                            */
#define CHANNEL_BAUDRATE     500000         /* Master channel baud rate                                     */
#define Wake_Up_Can_ID              0x21

/* Interrupt Service Routine priorities for Master and Slave SPI communication */
#define ISR_PRIORITY_MASTER_TX      40
#define ISR_PRIORITY_MASTER_RX      41

/*********************************************************************************************************************/
/*-------------------------------------------------Global variables--------------------------------------------------*/
/*********************************************************************************************************************/
qspiComm qspi;

/*********************************************************************************************************************/
/*------------------------------------------------Function Prototypes------------------------------------------------*/
/*********************************************************************************************************************/
void init_QSPI3_Master(void);
void initQSPI3MasterChannel(void);
void initQSPI3MasterBuffers(void);

/*********************************************************************************************************************/
/*----------------------------------------------Function Implementations---------------------------------------------*/
/*********************************************************************************************************************/
IFX_INTERRUPT(masterTxISR, 0, ISR_PRIORITY_MASTER_TX);                  /* SPI Master ISR for transmit data         */
IFX_INTERRUPT(masterRxISR, 0, ISR_PRIORITY_MASTER_RX);                  /* SPI Master ISR for receive data          */

void masterTxISR()
{
    IfxCpu_enableInterrupts();
    IfxQspi_SpiMaster_isrTransmit(&qspi.spiMaster);
}

void masterRxISR()
{
    IfxCpu_enableInterrupts();
    IfxQspi_SpiMaster_isrReceive(&qspi.spiMaster);
}


/* QSPI Master initialization
 * This function initializes the QSPI2 module in Master mode.
 */
void initQSPI3Master(void)
{
    IfxQspi_SpiMaster_Config spiMasterConfig;                           /* Define a Master configuration            */

    IfxQspi_SpiMaster_initModuleConfig(&spiMasterConfig, QSPI3_MASTER); /* Initialize it with default values        */

    spiMasterConfig.base.mode = SpiIf_Mode_master;                      /* Configure the mode                       */

    /* Select the port pins for communication */
    const IfxQspi_SpiMaster_Pins qspi3MasterPins = {
        &IfxQspi3_SCLK_P02_7_OUT, IfxPort_OutputMode_pushPull,          /* SCLK Pin                       (CLK)     */
        &IfxQspi3_MTSR_P02_6_OUT, IfxPort_OutputMode_pushPull,          /* MasterTransmitSlaveReceive pin (MOSI)    */
        &IfxQspi3_MRSTA_P02_5_IN, IfxPort_InputMode_pullDown,           /* MasterReceiveSlaveTransmit pin (MISO)    */
        IfxPort_PadDriver_cmosAutomotiveSpeed3                          /* Pad driver mode                          */
    };
    spiMasterConfig.pins = &qspi3MasterPins;                            /* Assign the Master's port pins            */

    /* Set the ISR priorities and the service provider */
    spiMasterConfig.base.txPriority = ISR_PRIORITY_MASTER_TX;
    spiMasterConfig.base.rxPriority = ISR_PRIORITY_MASTER_RX;
    spiMasterConfig.base.isrProvider = IfxSrc_Tos_cpu4;

    /* Initialize the QSPI Master module */
    IfxQspi_SpiMaster_initModule(&qspi.spiMaster, &spiMasterConfig);
}

/* QSPI Master channel initialization
 * This function initializes the QSPI2 Master channel.
 */
void initQSPI3MasterChannel(void)
{
    IfxQspi_SpiMaster_ChannelConfig spiMasterChannelConfig;             /* Define a Master Channel configuration    */

    /* Initialize the configuration with default values */
    IfxQspi_SpiMaster_initChannelConfig(&spiMasterChannelConfig, &qspi.spiMaster);

    spiMasterChannelConfig.base.baudrate = CHANNEL_BAUDRATE;     /* Set SCLK frequency to 1 MHz              */
    spiMasterChannelConfig.base.mode.clockPolarity = SpiIf_ClockPolarity_idleLow;
    spiMasterChannelConfig.base.mode.shiftClock = SpiIf_ShiftClock_shiftTransmitDataOnLeadingEdge;

    /* Select the port pin for the Chip Select signal */
    const IfxQspi_SpiMaster_Output qspi3SlaveSelect = {                 /* QSPI2 Master selects the QSPI3 Slave     */
        &IfxQspi3_SLSO0_P02_4_OUT, IfxPort_OutputMode_pushPull,         /* Slave Select port pin (CS)               */
        IfxPort_PadDriver_cmosAutomotiveSpeed1                          /* Pad driver mode                          */
    };
    spiMasterChannelConfig.sls.output = qspi3SlaveSelect;

    /* Initialize the QSPI Master channel */
    IfxQspi_SpiMaster_initChannel(&qspi.spiMasterChannel, &spiMasterChannelConfig);
}

/* QSPI Master SW buffer initialization
 * This function initializes SW buffers the Master uses.
 */
void initQSPI3MasterBuffers(void)
{
    for (uint8 i = 0; i < BUFFER_SIZE; i++)
    {
        qspi.spiBuffers.spiMasterTxBuffer[i] = 0;                     /* Clear TX Buffer                          */
        qspi.spiBuffers.spiMasterRxBuffer[i] = 0;                     /* Clear RX Buffer                          */
    }
}

/***********************************************************************************************************************
 * Function Name : init_QSPI3_Master
 * Description   : Spi 初始化函数
 * Arguments     :
 * Return Value  :
 * Author        : Joe
 * Date          : 2022.10.13
 ***********************************************************************************************************************/
void init_QSPI3_Master(void)
{
    initQSPI3Master();
    initQSPI3MasterChannel();
    initQSPI3MasterBuffers();
}

/***********************************************************************************************************************
 * Function Name : Read_Reg_TJA1145
 * Description   : 1145 Spi读函数
 * Arguments     : Reg_Addr   寄存器地址
 * Return Value  :
 * Author        : Joe
 * Date          : 2022.10.13
 ***********************************************************************************************************************/
void Read_Reg_TJA1145(uint8 Reg_Addr)
{
    while(IfxQspi_SpiMaster_getStatus(&qspi.spiMasterChannel) == SpiIf_Status_busy)
    {   /* Wait until the previous communication has finished, if any */
    }

    uint8 Instruction = ((Reg_Addr << 1) | 0x01);
    qspi.spiBuffers.spiMasterTxBuffer[0] = Instruction;              /* Fill TX Master Buffer with pattern       */
    qspi.spiBuffers.spiMasterTxBuffer[1] = 0xEF;                     /* Fill TX Master Buffer with pattern       */

    /* Send a data stream through the SPI Master */
    IfxQspi_SpiMaster_exchange(&qspi.spiMasterChannel, &qspi.spiBuffers.spiMasterTxBuffer[0], &qspi.spiBuffers.spiMasterRxBuffer[0], BUFFER_SIZE);
    waitTime(IfxStm_getTicksFromMilliseconds(BSP_DEFAULT_TIMER, 100));
    IfxQspi_SpiMaster_exchange(&qspi.spiMasterChannel, NULL_PTR, &qspi.spiBuffers.spiMasterRxBuffer[0], BUFFER_SIZE);
}

/***********************************************************************************************************************
 * Function Name : Write_Reg_TJA1145
 * Description   : 1145 Spi写函数
 * Arguments     : Reg_Addr   寄存器地址
 * Arguments     : Reg_Value  要写入的值
 * Return Value  :
 * Author        : Joe
 * Date          : 2022.10.13
 ***********************************************************************************************************************/
void Write_Reg_TJA1145(uint8 Reg_Addr, uint8 Reg_Value)
{
    while(IfxQspi_SpiMaster_getStatus(&qspi.spiMasterChannel) == SpiIf_Status_busy)
    {   /* Wait until the previous communication has finished, if any */
    }

    uint8 Instruction = ((Reg_Addr << 1) & 0xFE);
    qspi.spiBuffers.spiMasterTxBuffer[0] = Instruction;              /* Fill TX Master Buffer with pattern       */
    qspi.spiBuffers.spiMasterTxBuffer[1] = Reg_Value;                /* Fill TX Master Buffer with pattern       */

    /* Send a data stream through the SPI Master */
//    IfxQspi_SpiMaster_exchange(&qspi.spiMasterChannel, &qspi.spiBuffers.spiMasterTxBuffer[0], &qspi.spiBuffers.spiMasterRxBuffer[0], BUFFER_SIZE);
    IfxQspi_SpiMaster_exchange(&qspi.spiMasterChannel, &qspi.spiBuffers.spiMasterTxBuffer[0], NULL_PTR, BUFFER_SIZE);
    waitTime(IfxStm_getTicksFromMilliseconds(BSP_DEFAULT_TIMER, 100));
//    IfxQspi_SpiMaster_exchange(&qspi.spiMasterChannel, &qspi.spiBuffers.spiMasterTxBuffer[0], &qspi.spiBuffers.spiMasterRxBuffer[0], BUFFER_SIZE);
}

/* Reserved Function for test */
//void Write_Test(uint8 Reg_Addr, uint8 Reg_Value)
//{
//    Write_Reg_TJA1145(Reg_Addr,Reg_Value);
//    Read_Reg_TJA1145(Reg_Addr);
//    Ifx_Console_print("%X: %X\r\n\n", Reg_Addr, qspi.spiBuffers.spiMasterRxBuffer[1]);
//}

/***********************************************************************************************************************
 * Function Name : Chip_1145_Config
 * Description   : 配置1145芯片
 * Arguments     :
 * Return Value  :
 * Author        : Joe
 * Date          : 2022.10.13
 ***********************************************************************************************************************/
void Chip_1145_Config(void)
{
    init_QSPI3_Master();
    Write_Reg_TJA1145(0x01,0x07);
    Write_Reg_TJA1145(0x0A,0x00);
    Read_Reg_TJA1145(0x7E);
//    Ifx_Console_print("0x7E: %X\r\n\n", qspi.spiBuffers.spiMasterRxBuffer[1]);

    uint8 zyx = 0;
    Read_Reg_TJA1145(0x61);
    zyx = qspi.spiBuffers.spiMasterRxBuffer[1];
    Write_Reg_TJA1145(0x61,zyx);
    Read_Reg_TJA1145(0x63);
    zyx = qspi.spiBuffers.spiMasterRxBuffer[1];
    Write_Reg_TJA1145(0x63,zyx);
    Read_Reg_TJA1145(0x64);
    zyx = qspi.spiBuffers.spiMasterRxBuffer[1];
    Write_Reg_TJA1145(0x64,zyx);
    Write_Reg_TJA1145(0x23,0x01);
    Write_Reg_TJA1145(0x2B,0x00);
    Write_Reg_TJA1145(0x2C,0x00);
    Write_Reg_TJA1145(0x2D,0x00);
    Write_Reg_TJA1145(0x2E,0x00);
    Write_Reg_TJA1145(0x27,Wake_Up_Can_ID);
    Write_Reg_TJA1145(0x28,Wake_Up_Can_ID>>8);
    Write_Reg_TJA1145(0x29,Wake_Up_Can_ID>>16);
    Write_Reg_TJA1145(0x2A,Wake_Up_Can_ID>>24);
    Write_Reg_TJA1145(0x2F,0x80);
    Write_Reg_TJA1145(0x26,0x05);
    Write_Reg_TJA1145(0x20,0x31);
    Write_Reg_TJA1145(0x01,0x01);
}

/***********************************************************************************************************************
 * Function Name : Detect_Chip_1145_Wake_Up
 * Description   : 检测1145唤醒
 * Arguments     :
 * Return Value  :
 * Author        : Joe
 * Date          : 2022.10.13
 ***********************************************************************************************************************/
void Detect_Chip_1145_Wake_Up(void)
{
    Read_Reg_TJA1145(0x01);
    if(qspi.spiBuffers.spiMasterRxBuffer[1] == 4){
        Write_Reg_TJA1145(0x01,0x07);
    }
}


